/*
 * Copyright 2024 KylinSoft Co., Ltd.
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <https://www.gnu.org/licenses/>.
 */

#include "audioconfig.h"

#include <logger.h>

#include <cstring>

#include "_audioconfig.h"
#include "_audiodatastream.h"

std::string audioConfigToJson(const _AudioConfig &config) {
    Json::Value root;

    // 设置输入来源类型
    switch (config.input_source) {
        case _AudioConfig::DEFAULT_MIC:
            root["input_source"] = _AudioConfig::DEFAULT_MIC;
            break;
        case _AudioConfig::STREAM:
            root["input_source"] = _AudioConfig::STREAM;
            break;
        case _AudioConfig::PCM_FILE:
            root["input_source"] = _AudioConfig::PCM_FILE;
            root["pcm_file_path"] = config.pcm_file_path;
            break;
        case _AudioConfig::RAW_DATA:
            root["input_source"] = _AudioConfig::RAW_DATA;
            root["raw_data"] = util::base64Encode(config.raw_data.data,
                                                  config.raw_data.length);
            break;
        default:
            LOGE("未知输入源");
            break;
    }

    // 设置输出目标类型
    switch (config.output_target) {
        case _AudioConfig::DEFAULT_SPEAKER:
            root["output_target"] = _AudioConfig::DEFAULT_SPEAKER;
            break;
        case _AudioConfig::PCM_OUTPUT_FILE:
            root["output_target"] = _AudioConfig::PCM_OUTPUT_FILE;
            root["output_pcm_file_path"] =
                std::string(config.output_pcm_file_path);
            break;
        case _AudioConfig::PCM_DATA:
            root["output_target"] = _AudioConfig::PCM_DATA;
            break;
        default:
            LOGE("未知输出源");
            break;
    }

    // 设置采样率
    root["sample_rate"] = config.sample_rate;

    // 设置输出音频格式
    root["speed"] = config.speed;
    root["volume"] = config.volume;
    root["pitch"] = config.pitch;

    // 序列化 JSON 对象
    Json::FastWriter writer;
    return writer.write(root);
}

AudioConfig *audio_config_create_default() {
    AudioConfig *config = new _AudioConfig();
    // new出来的对象不要memset；联合体赋值是为了方便以后加数据
    config->input_source = _AudioConfig::DEFAULT_MIC;
    config->output_target = _AudioConfig::DEFAULT_SPEAKER;
    config->pcm_file_path = nullptr;
    config->raw_data.data = nullptr;
    config->raw_data.length = 0;
    config->output_pcm_file_path = nullptr;
    config->sample_rate = 16000;
    config->speed = 50;
    config->volume = 50;
    config->pitch = 50;
    return config;
}

AudioConfig *
audio_config_create_continuous_audio_input_from_default_microphone() {
    AudioConfig *config = audio_config_create_default();
    config->input_source = _AudioConfig::DEFAULT_MIC;
    return config;
}

AudioConfig *audio_config_create_continuous_audio_input_from_audio_data_stream(
    AudioDataStream *audio_data_stream) {
    AudioConfig *config = audio_config_create_default();
    config->input_source = _AudioConfig::STREAM;
    config->audio_data_stream = audio_data_stream;
    return config;
}

AudioConfig *audio_config_create_once_audio_input_from_pcm_file(
    const char *pcm_file) {
    AudioConfig *config = audio_config_create_default();
    config->input_source = _AudioConfig::PCM_FILE;
    config->pcm_file_path = strdup(pcm_file);
    return config;
}

AudioConfig *audio_config_create_once_audio_input_from_pcm_data(
    const uint8_t *audio_data, uint32_t data_length) {
    AudioConfig *config = audio_config_create_default();
    config->input_source = _AudioConfig::RAW_DATA;
    config->raw_data.data = audio_data;
    config->raw_data.length = data_length;
    return config;
}

AudioConfig *audio_config_create_audio_output_from_pcm_data() {
    AudioConfig *config = audio_config_create_default();
    config->output_target = _AudioConfig::PCM_DATA;
    return config;
}

// 使用系统默认扬声器作为音频输出
AudioConfig *audio_config_create_audio_output_from_default_speaker() {
    AudioConfig *config = audio_config_create_default();
    config->output_target = _AudioConfig::DEFAULT_SPEAKER;
    return config;
}

// 使用pcm文件作为音频输出
AudioConfig *audio_config_create_audio_output_from_pcm_file_name(
    const char *pcm_file) {
    AudioConfig *config = audio_config_create_default();
    config->output_target = _AudioConfig::PCM_OUTPUT_FILE;
    config->output_pcm_file_path = pcm_file;
    return config;
}

void audio_config_destroy(AudioConfig **config) {
    if (config == nullptr) {
        return;
    }
    delete *config;
    *config = nullptr;
}

void audio_config_set_input_audio_rate(AudioConfig *config, int rate) {
    if (config == nullptr) {
        return;
    }
    config->sample_rate = rate;
}

void audio_config_set_output_audio_speed(AudioConfig *config, int speed) {
    if (config == nullptr) {
        return;
    }
    config->speed = speed;
}

void audio_config_set_output_audio_volume(AudioConfig *config, int volume) {
    if (config == nullptr) {
        return;
    }
    config->volume = volume;
}

void audio_config_set_output_audio_pitch(AudioConfig* config, int pitch){
    if (config == nullptr) {
        return;
    }
    config->pitch = pitch;
}