/*
 * Copyright 2024 KylinSoft Co., Ltd.
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <https://www.gnu.org/licenses/>.
 */

#include "coreaispeechserviceproxy.h"

#include <string>
#include <vector>

#include "_speechmodelconfig.h"
#include "_speechrecognitionresult.h"
#include "_speechsynthesisresult.h"
#include "coreaispeechserver.h"
#include "logger.h"
#include "util.h"

namespace kyai {
namespace core {
namespace speech {

const char *objectPath = "/com/kylin/AiRuntime/CoreSpeech";
const char *interface = "com.kylin.AiRuntime.CoreSpeech";

const char *recognitionSpeechResultSignal = "RecognitionSpeechResult";

const char *synthesisResultSignal = "SynthesisResult";

_SpeechRecognitionResult dissectRecognitionResult(GVariant *parameters) {
    _SpeechRecognitionResult result{};

    GVariantIter iter;
    g_variant_iter_init(&iter, parameters);

    GVariant *reason = g_variant_iter_next_value(&iter);
    if (reason != nullptr) {
        int *pReason = (int *)g_variant_get_data(reason);
        if (pReason != nullptr) {
            result.reason = SpeechResultReason(*pReason);
        } else {
            LOGE("recognition result error: result is nullptr!");
        }
        g_variant_unref(reason);
    } else {
        LOGE("recognition result error: result is nullptr!");
    }

    GVariant *text = g_variant_iter_next_value(&iter);
    if (text != nullptr) {
        char *pText = (char *)g_variant_get_data(text);
        if (pText != nullptr) {
            result.text = pText;
        } else {
            LOGE("recognition result error: result is nullptr!");
        }
        g_variant_unref(text);
    } else {
        LOGE("recognition result error: result is nullptr!");
    }

    GVariant *speakerId = g_variant_iter_next_value(&iter);
    if (speakerId != nullptr) {
        int *pSpeakerId = (int *)g_variant_get_data(speakerId);
        if (pSpeakerId != nullptr) {
            result.speakerId = *pSpeakerId;
        } else {
            LOGE("recognition result error: result is nullptr!");
        }
        g_variant_unref(speakerId);
    } else {
        LOGE("recognition result error: result is nullptr!");
    }

    GVariant *errorCode = g_variant_iter_next_value(&iter);
    if (errorCode != nullptr) {
        int *pErrorCode = (int *)g_variant_get_data(errorCode);
        if (pErrorCode != nullptr) {
            result.errorCode = *pErrorCode;
        } else {
            LOGE("recognition result error: result is nullptr!");
        }
        g_variant_unref(errorCode);
    } else {
        LOGE("recognition result error: result is nullptr!");
    }

    GVariant *errorMessage = g_variant_iter_next_value(&iter);
    if (errorMessage != nullptr) {
        char *pErrorMessage = (char *)g_variant_get_data(errorMessage);
        if (pErrorMessage != nullptr) {
            result.errorMessage = pErrorMessage;
        } else {
            LOGE("recognition result error: result is nullptr!");
        }
        g_variant_unref(errorMessage);
    } else {
        LOGE("recognition result error: result is nullptr!");
    }
    return result;
}

_SpeechSynthesisResult dissectSynthesisResult(GVariant *parameters) {
    _SpeechSynthesisResult result{};

    GVariantIter iter;
    g_variant_iter_init(&iter, parameters);

    // 获取字节数组
    GVariant *audioData = g_variant_iter_next_value(&iter);
    if (audioData != nullptr) {
        gchar *pAudioData = (gchar *)g_variant_get_data(audioData);
        if (pAudioData != nullptr) {
            std::string encode = pAudioData;
            result.audioData = util::base64Decode(encode);
        } else {
            LOGE("recognition result error: result is nullptr!");
        }
        g_variant_unref(audioData);
    } else {
        LOGE("recognition result error: result is nullptr!");
    }

    GVariant *reason = g_variant_iter_next_value(&iter);
    if (reason != nullptr) {
        int *pReason = (int *)g_variant_get_data(reason);
        if (pReason != nullptr) {
            result.reason = SpeechResultReason(*pReason);
        } else {
            LOGE("recognition result error: result is nullptr!");
        }
        g_variant_unref(reason);
    } else {
        LOGE("recognition result error: result is nullptr!");
    }

    // 获取整数
    GVariant *audioFormat = g_variant_iter_next_value(&iter);
    if (audioFormat != nullptr) {
        char *pAudioFormat = (char *)g_variant_get_data(audioFormat);
        if (pAudioFormat != nullptr) {
            result.audioFormat = pAudioFormat;
        } else {
            LOGE("recognition result error: result is nullptr!");
        }
        g_variant_unref(audioFormat);
    } else {
        LOGE("recognition result error: result is nullptr!");
    }

    GVariant *audioRate = g_variant_iter_next_value(&iter);
    if (audioRate != nullptr) {
        int *pAudioRate = (int *)g_variant_get_data(audioRate);
        if (pAudioRate != nullptr) {
            result.audioRate = *pAudioRate;
        } else {
            LOGE("recognition result error: result is nullptr!");
        }
        g_variant_unref(audioRate);
    } else {
        LOGE("recognition result error: result is nullptr!");
    }

    GVariant *audioChannel = g_variant_iter_next_value(&iter);
    if (audioChannel != nullptr) {
        int *pAudioChannel = (int *)g_variant_get_data(audioChannel);
        if (pAudioChannel != nullptr) {
            result.audioChannel = *pAudioChannel;
        } else {
            LOGE("recognition result error: result is nullptr!");
        }
        g_variant_unref(audioChannel);
    } else {
        LOGE("recognition result error: result is nullptr!");
    }

    GVariant *errorCode = g_variant_iter_next_value(&iter);
    if (errorCode != nullptr) {
        int *pErrorCode = (int *)g_variant_get_data(errorCode);
        if (pErrorCode != nullptr) {
            result.errorCode = *pErrorCode;
        } else {
            LOGE("recognition result error: result is nullptr!");
        }
        g_variant_unref(errorCode);
    } else {
        LOGE("recognition result error: result is nullptr!");
    }

    GVariant *errorMessage = g_variant_iter_next_value(&iter);
    if (errorMessage != nullptr) {
        char *pErrorMessage = (char *)g_variant_get_data(errorMessage);
        if (pErrorMessage != nullptr) {
            result.errorMessage = pErrorMessage;
        } else {
            LOGE("recognition result error: result is nullptr!");
        }
        g_variant_unref(errorMessage);
    } else {
        LOGE("recognition result error: result is nullptr!");
    }

    return result;
}

void CoreAiSpeechServiceProxy::setRecognitionResultCallback(
    SpeechRecognitionResultCallback callback, void *userData) {
    recognitionCallback_ = callback;
    recognitionUserData_ = userData;
}

void CoreAiSpeechServiceProxy::setSynthesizerCallback(
    SpeechSynthesisResultCallback callback, void *userData) {
    synthesizerCallback_ = callback;
    synthesizerUserData_ = userData;
}

CoreAiSpeechServiceProxy::~CoreAiSpeechServiceProxy() {
    if (not CoreAiSpeechServer::getInstance().available()) {
        g_printerr(
            "Error creating core speech server proxy: Server proxy "
            "connection is unavailable.\n");
        return;
    }

    auto *connection = CoreAiSpeechServer::getInstance().getConnection();
    if (recognitionSubscriptionId_ != 0) {
        // 取消订阅
        g_dbus_connection_signal_unsubscribe(connection,
                                             recognitionSubscriptionId_);
    }
    if (synthesizerSubscriptionId_ != 0) {
        // 取消订阅
        g_dbus_connection_signal_unsubscribe(connection,
                                             synthesizerSubscriptionId_);
    }
}

int CoreAiSpeechServiceProxy::connectToServer() {
    if (!CoreAiSpeechServer::getInstance().available()) {
        g_printerr(
            "Error creating core speech server proxy: Server proxy "
            "connection is unavailable.\n");
        return -1;
    }

    GError *error = nullptr;
    auto *connection = CoreAiSpeechServer::getInstance().getConnection();
    proxy_ = ai_runtime_core_speech_service_proxy_new_sync(
        connection, G_DBUS_PROXY_FLAGS_NONE, nullptr, objectPath, nullptr,
        &error);
    if (proxy_ == nullptr) {
        g_printerr("Error creating text processor proxy %s: %s\n", objectPath,
                   error->message);
        g_error_free(error);
        return -1;
    }

    return 1;
}

int CoreAiSpeechServiceProxy::initRecognizer(int &errorCode,
                                             const char *config) {
    int sessionId = -1;
    bool success = ai_runtime_core_speech_service_call_init_recognizer_sync(
        proxy_, config, &sessionId, &errorCode, nullptr, nullptr);
    if (not success) {
        return -1;
    }
    if (sessionId == -1) {
        return -1;
    }
    connectSpeechRecognitionResultSignals(sessionId);
    return sessionId;
}

int CoreAiSpeechServiceProxy::initSynthesizer(int &errorCode,
                                              const char *config) {
    int sessionId = -1;
    bool success = ai_runtime_core_speech_service_call_init_synthesizer_sync(
        proxy_, config, &sessionId, &errorCode, nullptr, nullptr);
    if (not success) {
        return -1;
    }
    if (sessionId == -1) {
        return -1;
    }

    connectSpeechSynthesisResultSignals(sessionId);
    return sessionId;
}

void CoreAiSpeechServiceProxy::destroyRecognizer(int sessionId) const {
    ai_runtime_core_speech_service_call_destroy_recognizer_sync(
        proxy_, sessionId, nullptr, nullptr);
}

void CoreAiSpeechServiceProxy::destroySynthesizer(int sessionId) const {
    ai_runtime_core_speech_service_call_destroy_synthesizer_sync(
        proxy_, sessionId, nullptr, nullptr);
}

void CoreAiSpeechServiceProxy::setRecognizerAudioConfig(
    int sessionId, const std::string &audioConfig) {
    ai_runtime_core_speech_service_call_set_recognizer_audio_config_sync(
        proxy_, sessionId, audioConfig.c_str(), nullptr, nullptr);
    LOGD("setRecognizerAudioConfig:");
}

void CoreAiSpeechServiceProxy::startContinuousRecognition(int sessionId) {
    ai_runtime_core_speech_service_call_start_continuous_recognition(
        proxy_, sessionId, nullptr, nullptr, nullptr);
    LOGD("startContinuousRecognition.");
}

void CoreAiSpeechServiceProxy::setSynthesizerAudioConfig(
    int sessionId, const std::string &audioConfig) {
    bool ret =
        ai_runtime_core_speech_service_call_set_synthesizer_audio_config_sync(
            proxy_, sessionId, audioConfig.c_str(), nullptr, nullptr);
    if (!ret) {
        std::fprintf(
            stderr,
            "CoreSpeechServiceProxy: error to set synthesizer audio config\n");
    }
}

void CoreAiSpeechServiceProxy::synthesizeTextOnce(int sessionId,
                                                  const char *text,
                                                  uint32_t textLength) {
    ai_runtime_core_speech_service_call_synthesize_text_once(
        proxy_, sessionId, text, textLength, nullptr, nullptr, nullptr);
}

void CoreAiSpeechServiceProxy::synthesizeTextContinuous(int sessionId,
                                                        const char *text,
                                                        uint32_t textLength) {
    ai_runtime_core_speech_service_call_synthesize_text_continuous(
        proxy_, sessionId, text, textLength, nullptr, nullptr, nullptr);
}

int CoreAiSpeechServiceProxy::stopSpeaking(int sessionId) {
    int errorCode;
    GError *error = nullptr;
    bool ret = ai_runtime_core_speech_service_call_stop_speaking_sync(
        proxy_, sessionId, &errorCode, nullptr, &error);
    if (!ret) {
        LOGE("Error to stop speaking.", error->message);
        return AiSdkCommonErrorCode::AISDK_RUNTIME_ERROR;
    }
    return (int)errorCode;
}

void CoreAiSpeechServiceProxy::stopContinuousRecognition(int sessionId) {
    ai_runtime_core_speech_service_call_stop_continuous_recognition(
        proxy_, sessionId, nullptr, nullptr, nullptr);
    LOGI("stopContinuousRecognition.");
}
void CoreAiSpeechServiceProxy::continuousRecognitionWriteAudioData(
    int sessionId, const uint8_t *data, uint32_t length) {
    const char *ptr = reinterpret_cast<const char *>(data);
    std::string audioData(ptr, length);
    std::string encode = util::base64Encode(audioData);
    ai_runtime_core_speech_service_call_continuous_recognition_write_audio_data(
        proxy_, sessionId, encode.c_str(), nullptr, nullptr, nullptr);
    // LOGI("continuousRecognitionWriteAudioData.");
}

void CoreAiSpeechServiceProxy::recognizeOnce(int sessionId) {
    ai_runtime_core_speech_service_call_recognize_once(
        proxy_, sessionId, nullptr, nullptr, nullptr);
    LOGD("recognizeOnce.");
}

void CoreAiSpeechServiceProxy::connectSpeechRecognitionResultSignals(
    int sessionId) {
    auto *connection = CoreAiSpeechServer::getInstance().getConnection();

    std::string interfaceName = interface + std::to_string(sessionId);

    recognitionSubscriptionId_ = g_dbus_connection_signal_subscribe(
        connection, nullptr, interfaceName.c_str(),
        recognitionSpeechResultSignal, objectPath, nullptr,
        G_DBUS_SIGNAL_FLAGS_NONE, onRecognitionResultCallback, this, nullptr);
}

void CoreAiSpeechServiceProxy::connectSpeechSynthesisResultSignals(
    int sessionId) {
    auto *connection = CoreAiSpeechServer::getInstance().getConnection();
    std::string interfaceName = interface + std::to_string(sessionId);

    synthesizerSubscriptionId_ = g_dbus_connection_signal_subscribe(
        connection, nullptr, interfaceName.c_str(), synthesisResultSignal,
        objectPath, nullptr, G_DBUS_SIGNAL_FLAGS_NONE, onSynthesisResult, this,
        nullptr);
}

void CoreAiSpeechServiceProxy::onRecognitionResultCallback(
    GDBusConnection *connection, const gchar *senderName,
    const gchar *objectPath, const gchar *interfaceName,
    const gchar *signalName, GVariant *parameters, gpointer userData) {
    auto *proxy = static_cast<CoreAiSpeechServiceProxy *>(userData);
    if (proxy->recognitionCallback_ != nullptr) {
        _SpeechRecognitionResult result = dissectRecognitionResult(parameters);
        proxy->recognitionCallback_(&result, proxy->recognitionUserData_);
    } else {
        LOGE("Recognizing callback is not set!");
    }
}

void CoreAiSpeechServiceProxy::onSynthesisResult(
    GDBusConnection *connection, const gchar *senderName,
    const gchar *objectPath, const gchar *interfaceName,
    const gchar *signalName, GVariant *parameters, gpointer userData) {
    auto *proxy = static_cast<CoreAiSpeechServiceProxy *>(userData);
    if (!proxy->synthesizerCallback_) {
        std::fprintf(
            stderr,
            "CoreSpeechServiceProxy: Synthesizer callback is not set!\n");
        return;
    }

    auto result = dissectSynthesisResult(parameters);
    proxy->synthesizerCallback_(&result, proxy->synthesizerUserData_);
}

}  // namespace speech
}  // namespace core
}  // namespace kyai
